/*******************************************************************************
 *  osinfo.c
 *  Program to display current max's min's of system variables.
 *
 ******************************************************************************/
#ifdef __USAGE
%C [ -w watermark -s sleep_period -p priority -b browser -f logfile ]

watermark:    If a parameter goes over this watermark (if a high parameter) or
              under (if a low) then a 'water hit' counter will be incremented.
              Default is 20%.
              This makes 20% of maximum for a low water mark and (100-20)=80%
              for a high water mark.

sleep_period: Seconds to sleep between polls. Default is 3.

priority:     Priority to run at. Default is current priority.

browser:      File browser to use. e.g. /bin/less

logfile:      Logfile created by dumplog.

e.g.
%C -h 10 -s 1 -p 20 -b ezv -f dumplogfile
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/osinfo.h>
#include <sys/psinfo.h>
#include <sys/fd.h>
#include <sys/name.h>
#include <sys/sidinfo.h>
#include <sys/irqinfo.h>
#include <sys/timers.h>
#include <i86.h>
#include <time.h>
#include <sys/proc_msg.h>
#include <sys/kernel.h>
#include <sys/qnxterm.h>
#include <sys/stat.h>
#include <signal.h>
#include <setjmp.h>
#include <process.h>
#include <sys/wait.h>

int qnx_timer_query(pid_t, timer_t);
time_t get_time(), get_log_time();

/* globals for display */

#define MAX_LINE_ITEMS 11
struct sys_data {
        int current;         /* currently read value */
        int highlow;         /* the highest or lowest value read */
        int max;             /* the maximum value of this parameter */
        int water;           /* a mark at which the value becomes critical */
        time_t ltime;        /* the last time the highlow value was changed */
        int row;             /* the row to print at */
        int col;             /* the column to print at */
        int type;            /* 1 = show max, 0 = no max display */
        int water_hits;      /* the number of times we've hit water mark */
       };

struct sys_data data[MAX_LINE_ITEMS];

/* data arranged as follows ...
     0 = system memory
     1 = proc heap memory
     2 = proc entries (total of vids, pids and proxs )
     3 = executables
     4 = vids
     5 = proxies
     6 = file descriptors
     7 = names
     8 = timers
     9 = interrupt handlers
    10 = sessions
*/

/* default parameters */
int highwater   = 20;  /* percentage */
int poll_period = 3;
int priority    = -1;

char filename[255];
char browser[255];
int  check_dump;
time_t dump_lasttime = 0;
time_t starttime;

int no_display = 0;

pid_t  mypid;
struct _osinfo o;

#define NORMAL_LABEL  TERM_CYAN 
#define BLINK_FIELD   TERM_WHITE | TERM_HILIGHT | TERM_BLINK
#define DATA_FIELD    TERM_WHITE | TERM_NORMAL

volatile struct _timesel far *t_ptr;

sigjmp_buf env;

void breakit(int signo) {
   term_cur(24,0);
   exit(-1);
   }

void timeout(int signo) {
   siglongjmp(env, 1);
}

main(argc,argv)
int argc;
char *argv[];
{
   int     c;
   static timer_t           timer;  /* timer id */
   static struct itimercb   timerc;	/* timer control block */
   static struct itimerspec timers;	/* timer value specifier */
   char buffer[512];
   int ret_val;
   int status;
   pid_t wpid;

   while ( ( c = getopt( argc, argv, "w:p:s:f:b:" ) ) != -1 ) {
         switch(c) {
            case 'w' : highwater = atoi(optarg);
                       break;
            case 'p' : priority = atoi(optarg);
                       break;
            case 's' : poll_period = atoi(optarg);
                       break;
            case 'b' : strcpy(browser,optarg);
                       break;
            case 'f' : strcpy(filename,optarg);
                       break;
            }
         }

   mypid = getpid();

   timerc.itcb_event.evt_value = SIGALRM;
   if ((timer = mktimer(TIMEOFDAY, DELIVERY_SIGNALS, &timerc)) == -1) {
       fprintf(stderr, "%s (%s)", "unable to make timer", strerror(errno));
       exit(-1);
       }

   if ( priority >= 0 && priority <= 29 )
       if ( setprio( mypid, priority ) == -1 ) {
          fprintf(stderr,"Error setting priority. (%s)\n",strerror(errno));
          exit(-1);
          }

   if ( term_load() == -1 ) {
      fprintf(stderr, "term_load error: %s\n", strerror(errno));
      exit(-1);
      }

   signal(SIGINT,  breakit);
   signal(SIGALRM, timeout);

   setup_initial_data();
   display_screen();
   get_info();        /* get the data the first time through */

   ret_val = sigsetjmp(env, 1);

   while(1) {
      switch ( ret_val ) {
          case 0 : /* initial sigsetjmp .. do nothing */
                   break;
          default: /* got a call from siglongjmp(). means we got the signal */
                   get_info();
                   if ( !no_display )
                      update_screen();
                   break;
          }

      timers.it_value.tv_sec = (long)(poll_period);
      reltimer(timer, &timers, 0);

      switch ( no_display ) {
          case 1 : /* we're browsing a file, simply wait on child */
                   c = -1;   
                   wpid = wait( &status );
                   if ( wpid != -1 ) {
                      no_display = 0;
                      display_screen();
                      }
                   break;
          default: /* wait on a character */
                   c = getch();
                   break;
          }

      /* if we're down here then either wait() returned or getch returned */
      timers.it_value.tv_sec = 0;
	  reltimer(timer, &timers, 0);

      switch( c ) {
           case 'd':
           case 'D': /* check dump log */
                     dump_lasttime = get_log_time();
                     term_clear( TERM_CLS_SCRH );
                     term_flush();

                     switch ( spawnlp( P_NOWAIT, browser, browser,
                                          filename, NULL) ) {
                        case -1: /* couldn't do it */
                             term_printf( 24, 10, NORMAL_LABEL,
                                         "Couldn't load %s", strerror(errno));
                             break;
                        default: /* okay .. in browser */
                             no_display = 1;
                             break;
                        }
/*
                     sprintf(buffer,"%s %s",browser, filename);
                     system(buffer);
*/
                     break;
           default:  break;
           }
      }
}

setup_initial_data()
{
   int i;
   int row = 5;
   int col = 17;

   /* read osinfo and set highwater marks */
   qnx_osinfo( 0, &o );
   t_ptr = MK_FP(o.timesel,0);     /* get a pointer into system time */
   
   data[0].max = o.totmemk;      data[0].highlow = o.totmemk;     
   data[0].row = row;            data[0].col = col;  row++;
   data[1].max = o.proc_freemem; data[1].highlow = o.proc_freemem;
   data[1].row = row;            data[1].col = col;  row++; row++; row++; row++;
   data[2].max = o.num_procs;    data[2].highlow = 0;
   data[2].row = row;            data[2].col = col;  row++;
   data[3].max = o.num_procs;    data[3].highlow = 0;
   data[3].row = row;            data[3].col = col;  row++;
   data[4].max = o.num_procs;    data[4].highlow = 0;
   data[4].row = row;            data[4].col = col;  row++;
   data[5].max = o.num_procs;    data[5].highlow = 0;
   data[5].row = row;            data[5].col = col;  row++; row++;
   data[6].max = o.num_fds[2];   data[6].highlow = 0;
   data[6].row = row;            data[6].col = col;  row++;
   data[7].max = o.num_names;    data[7].highlow = 0;
   data[7].row = row;            data[7].col = col;  row++;
   data[8].max = o.num_timers;   data[8].highlow = 0;
   data[8].row = row;            data[8].col = col;  row++;
   data[9].max = o.num_handlers; data[9].highlow = 0;
   data[9].row = row;            data[9].col = col;  row++;
   data[10].max= o.num_sessions; data[10].highlow= 0;
   data[10].row= row;            data[10].col = col; row++;

   for ( i=0; i < MAX_LINE_ITEMS; i++ ) {
       if ( i == 0 || i == 1 ) /* memory stuff is a low */
          data[i].water = data[i].max * highwater / 100;
       else
          data[i].water = data[i].max * (100 - highwater) / 100;
       data[i].type = 1;   /* default */
       data[i].water_hits = 0;
       data[i].ltime = get_time();
       }
   data[1].type = data[3].type = data[4].type = data[5].type = 0;

   starttime = get_time();
}

display_screen()
{
int row = 0;
int col = 0;

term_clear( TERM_CLS_SCR );

term_printf(row, col, NORMAL_LABEL,
  "System Water Marks:  Node %d, HighWater= %d%%, Polling Every %d Second(s).", 
  getnid(), 100-highwater, poll_period);  row++;
term_printf(row, col, NORMAL_LABEL,
  "                     Analysis started on: %19.19s", 
  asctime(localtime(&starttime)) );    row++;
row++;

term_type(row, col, "Memory            current      low        max   when                 water hits", 0, NORMAL_LABEL ); row++;
term_type(row, col, "-------------------------------------------------------------------------------", 0, NORMAL_LABEL ); row++;
term_type(row, col, "Free Memory (k)", 0, NORMAL_LABEL ); row++;
term_type(row, col, "Proc (bytes)", 0, NORMAL_LABEL ); row++;
row++;
term_type(row, col, "Resource          current     high        max   when                 water hits", 0, NORMAL_LABEL ); row++;
term_type(row, col, "-------------------------------------------------------------------------------", 0, NORMAL_LABEL ); row++;
term_type(row, col, "Proc Entries", 0, NORMAL_LABEL ); row++;
term_type(row, col, "   executables", 0, NORMAL_LABEL ); row++;
term_type(row, col, "   vc's", 0, NORMAL_LABEL ); row++;
term_type(row, col, "   proxies", 0, NORMAL_LABEL ); row++;
row++;
term_type(row, col, "File Descriptors", 0, NORMAL_LABEL ); row++;
term_type(row, col, "Registered Names", 0, NORMAL_LABEL ); row++;
term_type(row, col, "Timers", 0, NORMAL_LABEL ); row++;
term_type(row, col, "IRQ Handlers", 0, NORMAL_LABEL ); row++;
term_type(row, col, "Sessions", 0, NORMAL_LABEL ); row++;
term_flush();
}

update_screen()
{
int i;
unsigned attr;

   for( i=0; i < MAX_LINE_ITEMS; i++ ) {
      attr = DATA_FIELD;
      switch ( data[i].type ) {
         case 1:  /* with a max field */
                  term_printf( data[i].row, data[i].col, attr,
                              "%8d %8d / %8d   %19.19s %7d", 
                               data[i].current, data[i].highlow, data[i].max,
                               asctime(localtime(&data[i].ltime)),
                               data[i].water_hits
                              );
                  break;
         case 0:  /* without a max field */
                  term_printf( data[i].row, data[i].col, attr,
                              "%8d %8d              %19.19s %7d", 
                               data[i].current, data[i].highlow,
                               asctime(localtime(&data[i].ltime)),
                               data[i].water_hits
                              );
                  break;
         }
      }

      if ( check_dump ) {
         term_printf( 21, 10, BLINK_FIELD,
                "New abnormal deaths since last check. Press 'D' for details.");
         }
      else {
         term_printf( 21, 10, NORMAL_LABEL,
                "            No abnormal deaths since last check.            ");
         }
      term_flush();
}

time_t
get_log_time()
{
struct stat buf;

if ( stat( filename, &buf ) != -1 ) 
   return(buf.st_mtime);
else
   return((time_t)0);
}

get_info()
{
int i;
pid_t id = 1; 
struct _psinfo p;

    for ( i=0; i < MAX_LINE_ITEMS; i++ )
        data[i].current = 0;

    qnx_osinfo( 0, &o );
    data[0].current = o.freememk; 
    data[1].current = o.proc_freemem; 

    while( ( id = qnx_psinfo( 0, id, &p, 0, 0 ) ) != -1 ) {
         if ( p.flags & _PPF_MID )
            data[5].current++;  /* proxies */
         else
         if ( p.flags & _PPF_VID )
            data[4].current++;  /* vc's */
         else
            data[3].current++;  /* executables */
         data[6].current += get_fds_for_this_process(id);   /* fd's */
         ++id;
         }
    data[2].current  = data[3].current + data[4].current + data[5].current;  /* total processes */
    data[7].current  = get_total_names();
    data[8].current  = get_total_timers();
    data[9].current  = get_total_handlers();
    data[10].current = get_total_sessions();

    /* check current highs */
    for ( i=0; i < MAX_LINE_ITEMS; i++ ) {
        switch ( i ) {
           case 0 :
           case 1 : /* low water mark stuff */
                    if ( data[i].current < data[i].highlow ) {
                       data[i].highlow = data[i].current; 
                       data[i].ltime = get_time();
                       }
                    if ( data[i].current < data[i].water ) { 
                       data[i].water_hits++;
                       }
                    break;
           default: /* everything else */
                    if ( data[i].current > data[i].highlow ) {
                       data[i].highlow = data[i].current; 
                       data[i].ltime = get_time();
                       }
                    if ( data[i].current > data[i].water ) { 
                       data[i].water_hits++;
                       }
                    break;
           }
        }

      if ( get_log_time() > dump_lasttime ) {
         check_dump = 1;
         }
      else {
         check_dump = 0;
         }
}

get_fds_for_this_process(pid_t id) {
int fd;
struct _fd_entry buf;
int totfd = 0;

    for ( fd=0; 
        ( fd = qnx_fd_query( 0, id, fd, &buf ) ) != -1;
          ++fd )
        totfd++;

    return(totfd);
}

get_total_names()
{
int id;
struct _nameinfo buf;
int totnames=0;

    for ( id=0; 
        ( id = qnx_name_query( 0, id, &buf ) ) != -1;
          ++id )
        totnames++;

    return(totnames);
}

get_total_sessions()
{
int id;
struct _sidinfo buf;
int totsessions=0;

    for ( id=0; 
        ( id = qnx_sid_query( 0, id, &buf ) ) != -1;
          ++id )
        totsessions++;

    return(totsessions);
}

get_total_handlers()
{
int id;
struct _irqinfo buf;
int tothandlers=0;

    for ( id=0; 
        ( id = qnx_hint_query( 0, id, &buf ) ) != -1;
          ++id )
        tothandlers++;

    return(tothandlers);
}

get_total_timers()
{
int id;
int tottimers=0;

    for ( id = 0; 
        ( id = qnx_timer_query( PROC_PID, id ) ) != -1;
          ++id )
        tottimers++;

    return(tottimers);
}

qnx_timer_query(proc_pid, timerid)
pid_t						proc_pid;
timer_t						timerid;
	{
	union stimer {
		struct _proc_timer		  s;
		struct _proc_timer_reply	r;
		} msg;

	msg.s.type = _PROC_TIMER;
	msg.s.subtype = _PROC_SUB_QUERY;
	msg.s.cookie = timerid;
	msg.s.flags = 0;
	msg.s.zero1 = 0;

	if(Send(proc_pid, &msg.s, &msg.r, sizeof(msg.s), sizeof(msg.r)) == -1)
		return(-1);

	if(msg.r.status) {
		return(-1);
		}

	return(msg.r.cookie);
	}

time_t
get_time()
{
return( t_ptr->seconds );
}

